﻿
using System;
using System.Drawing;
using System.Windows.Forms;

/* Interpolation by Chebychev polynomials */
/* www.mosismath.com */


namespace WindowsFormsApplication1
{

    public partial class MainWin : Form
    {
        public const int order = 6; // number of reference points
        public const int datapoints = 1000; // number of datapoints for the interpolation
        public const double deltaX = 0.5; // sample time 
        double[] t = new double[order];
        double[] c = new double[order];

        public class TMatrix
        {
            public double[,] a;
            public double[] y;
            public double[] x;
            public double[] t;

            public TMatrix(int size)
            {
                a = new double[size, size];
                y = new double[size];
                x = new double[size];
            }
        }

        TMatrix m;
        double[] yp = new double[datapoints];
        double[] xp = new double[datapoints];
        double[] y_k = new double[order];
        double[] x_k = new double[order];
 
        public class TGauss
        {
            public bool calcError;

            void SwitchRows(int n)
            {
                double tempD;
                int i, j;
                for (i = n; i <= maxOrder - 2; i++)
                {
                    for (j = 0; j <= maxOrder - 1; j++)
                    {
                        tempD = m.a[i, j];
                        m.a[i, j] = m.a[i + 1, j];
                        m.a[i + 1, j] = tempD;
                    }
                    tempD = m.y[i];
                    m.y[i] = m.y[i + 1];
                    m.y[i + 1] = tempD; 
                }
            }


            public int maxOrder = order;
            TMatrix m;

            public TGauss(int size, TMatrix mi)
            {
                maxOrder = size;
                m = mi;
            }

            public bool Eliminate()
            {
                int i, k, l;
                calcError = false;
                for (k = 0; k <= maxOrder - 2; k++)
                {
                    for (i = k; i <= maxOrder - 2; i++)
                    {
                        if (Math.Abs(m.a[i+1, i]) < 1e-8)
                        {
                            SwitchRows(i+1);
                        }
                        if (m.a[i + 1, k] != 0.0)
                        {
                            for (l = k + 1; l <= maxOrder - 1; l++)
                            {
                                if (!calcError)
                                {
                                    m.a[i + 1, l] = m.a[i + 1, l] * m.a[k, k] - m.a[k, l] * m.a[i + 1, k];
                                    if (m.a[i + 1, l] > 10E260)
                                    {
                                        m.a[i + 1, k] = 0;  // range overflow 
                                        calcError = true;
                                    }
                                }
                            }
                            m.y[i + 1] = m.y[i + 1] * m.a[k, k] - m.y[k] * m.a[i + 1, k];
                            m.a[i + 1, k] = 0;
                        }
                    }
                }
                return !calcError;
            }

            public void Solve()
            {
                int k, l;
                for (k = maxOrder - 1; k >= 0; k--)
                {
                    for (l = maxOrder - 1; l >= k; l--)
                    {
                        m.y[k] = m.y[k] - m.x[l] * m.a[k, l];
                    }
                    if (m.a[k, k] != 0)
                        m.x[k] = m.y[k] / m.a[k, k];
                    else
                        m.x[k] = 0;
                }
            }
        }

        TGauss gauss;

        public double CalcChebychev(double x, int order)
        {
            int i;
            double value;
            t[0] = 1;
            t[1] = x;
            for (i = 2; i < order; i++)
                t[i] = 2 * x * t[i - 1] - t[i - 2];
            t[0] = 0.5;
            value = c[0]/2.0;
            for (i = 1; i < order; i++)
                value = value + c[i] * t[i];
            return value;
        }

        
        public MainWin()
        {
            int i;
            m = new TMatrix(order);
            gauss = new TGauss(order, m);
            DataGridViewCell cell;

            InitializeComponent();

            dGPolynom.ColumnCount = order;
            dGPolynom.RowCount = 2;
 
             // some sample values for y
             y_k[0] = 0.5;
             y_k[1] = 11.7;
             y_k[2] = 14.8;
             y_k[3] = 4.0;
             y_k[4] = 2.2;
             y_k[5] = 0.2;

            // timestamp to y value
            for (i = 0; i < order; i++)
            {
                x_k[i] = deltaX * (double)(i);
            }
            cell = dGPolynom.Rows[0].HeaderCell;
            cell.Value = "y_k";
            cell = dGPolynom.Rows[1].HeaderCell;
            cell.Value = "x_k";
            for (i = 0; i < order; i++)
            {
                dGPolynom.Columns[i].Name = "a" + i.ToString();
                cell = dGPolynom[i, 0];
                cell.Value = Convert.ToString(y_k[i]);
                cell = dGPolynom[i, 1];
                cell.Value = Convert.ToString(x_k[i]);
            }
        }

        public void Interpolate(int maxPoints, double minTime, double maxTime)
        {
            int i;
            for (i = 0; i < maxPoints; i++)
            {
                yp[i] = 0;
                xp[i] = (double)(i) * (maxTime - minTime) / (double)maxPoints;
                yp[i] = CalcChebychev(xp[i], order);
            }
        }

        public void DrawGraph(int maxPoints, double minX, double maxX, bool doClear)
        {
            Point p1, p2;
            int j;
            double maxTime = -1.0;
            double maxVal = -1.0;
            double minVal = 1.0;
            int spacing = 20;
            double scalefactor;
            bool drawCeroline;
            for (j = 0; j < maxPoints - 1; j++)
            {
                if ((xp[j] >= minX) && (xp[j] <= maxX))
                {
                    if (maxTime < xp[j])
                        maxTime = xp[j];
                    if (maxVal < yp[j])
                        maxVal = yp[j];
                    if (minVal > yp[j])
                        minVal = yp[j];
                }
            }
            maxTime = maxTime * 1.05;
            maxTime = maxTime * 1.05;
            if (minVal < 0)
            {
                if (maxVal > 0)
                {
                    if (maxVal > Math.Abs(minVal))
                        scalefactor = pGraph.Height / maxVal / 2.2;
                    else
                        scalefactor = pGraph.Height / Math.Abs(minVal) / 2.2;
                }
                else
                    scalefactor = 5.0;
                drawCeroline = true;
            }
            else
            {
                scalefactor = pGraph.Height / (maxVal) / 1.1;
                drawCeroline = false;
            }
            p1 = new Point();
            p2 = new Point();
            Graphics g = pGraph.CreateGraphics();
            if (doClear)
                g.Clear(Color.White);
            Pen bluePen = new Pen(Color.Blue, 2);
            Pen redPen = new Pen(Color.Red, 2);
            Pen blackPen = new Pen(Color.Black, 2);
            if (drawCeroline)
            {
                p1.X = 0;
                p1.Y = (pGraph.Height / 2);
                p2.X = pGraph.Width;
                p2.Y = (pGraph.Height / 2);
                g.DrawLine(blackPen, p1, p2);
            }
            bluePen.Width = 1;
            if (maxTime > 0) // draw interpolated graph
            {
                for (j = 0; j < maxPoints - 1; j++)
                {
                    if ((xp[j] >= minX) && (xp[j] <= maxX))
                    {
                        p1.X = Convert.ToInt32(xp[j] * pGraph.Width / (double)maxTime);
                        p2.X = Convert.ToInt32(xp[j + 1] * pGraph.Width / (double)maxTime);
                        if (drawCeroline)
                        {
                            p1.Y = (pGraph.Height / 2) - Convert.ToInt32(Math.Round(yp[j] * scalefactor)) - spacing;
                            p2.Y = (pGraph.Height / 2) - Convert.ToInt32(Math.Round(yp[j + 1] * scalefactor)) - spacing;
                        }
                        else
                        {
                            p1.Y = (pGraph.Height) - Convert.ToInt32(Math.Round(yp[j] * scalefactor)) - spacing;
                            p2.Y = (pGraph.Height) - Convert.ToInt32(Math.Round(yp[j + 1] * scalefactor)) -spacing;
                        }
                        g.DrawLine(bluePen, p1, p2);
                    }
                }
                for (j = 0; j < order; j++) // mark reference points
                {
                    p1.X = Convert.ToInt32(x_k[j] * pGraph.Width / (double)maxTime);
                    if (drawCeroline)
                        p1.Y = (pGraph.Height / 2) - Convert.ToInt32(Math.Round(y_k[j] * scalefactor)) - spacing;
                    else
                        p1.Y = (pGraph.Height) - Convert.ToInt32(Math.Round(y_k[j] * scalefactor)) - spacing;
                    g.DrawEllipse(redPen, (int)(p1.X - 1.0), (int)(p1.Y - 1.0), 2, 2);
                }
            }
            g.Dispose();
        }


        private void button1_Click(object sender, EventArgs e)
        {
            int i, k;
            DataGridViewCell cell;

            for (i = 0; i < order; i++)
            {
                cell = dGPolynom[i, 0];
                if (cell.ToString() != "")
                {
                    y_k[i] = Convert.ToDouble(cell.Value);
                    cell = dGPolynom[i, 1];
                    x_k[i] = Convert.ToDouble(cell.Value);
                }
                else
                {
                    y_k[i] = 0.0;
                    x_k[i] = 0.0;
                }
                m.x[i] = 0.0;
            }

            // fill matrix upside down
            for (i = 0; i < order; i++)
            {
                CalcChebychev(x_k[order - 1 - i], order);
                m.y[i] = y_k[order - 1 - i];
                for (k = 0; k < order; k++)
                {
                    m.a[i, k] = t[k];
                }
            }
            if (gauss.Eliminate())
            {
                gauss.Solve();
                for (i = 0; i < order; i++)
                    c[i] = m.x[i];
                Interpolate(datapoints, x_k[0], x_k[order - 1]);
                DrawGraph(datapoints, x_k[0], x_k[order - 1], true);
            }
            else
                MessageBox.Show("Matrix calculattion", "Error", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
        }

        private void pGraph_Paint(object sender, PaintEventArgs e)
        {
            DrawGraph(datapoints, x_k[0], x_k[order - 1], false);
        }  
    }
}
